from .arguments import ConnectionArguments
from .openobd import *
from .exceptions import *
from .stream_handler import *
from .session import *
from .session_controller import *
from .session_token_handler import *
from .session_builder import *





import logging
from typing import Optional

from colorlog import ColoredFormatter

def initialize_logging(log_level: Optional[str]):
    if log_level is None or not isinstance(log_level, str):
        log_level = "INFO"

    handler = logging.StreamHandler()
    formatter = ColoredFormatter(
        "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(threadName)s: %(white)s%(message)s",
        log_colors={
            'DEBUG': 'cyan',
            'INFO': 'green',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'bold_red',
        }
    )
    handler.setFormatter(formatter)
    logging.basicConfig(handlers=[handler], level=log_level.upper())


from openobd.core.grpc_factory import NetworkGrpcFactory
from openobd.core.function_broker import OpenOBDFunctionBroker
from openobd.core.session import OpenOBDSession
from openobd.core.arguments import ConnectionArguments
from openobd_protocol.FunctionBroker import FunctionBrokerServices_pb2_grpc as grpcFunctionBrokerServices
from openobd_protocol.FunctionBroker.Messages.FunctionBroker_pb2 import FunctionUpdate, FunctionUpdateResponse, \
    FunctionCall, FunctionUpdateType, FunctionBrokerToken, FunctionSignature

from openobd.functions.function_executor import FunctionExecutor
import uuid



class BypassFunctionBrokerServicer(grpcFunctionBrokerServices.functionBrokerServicer):

    def __init__(self, arguments: ConnectionArguments, executor: FunctionExecutor, grpc_factory = None):
        self.arguments = arguments
        self.executor = executor
        self.grpc_factory = grpc_factory

    def runFunction(self, request: FunctionCall, **kwargs):
        '''Run function using the function executor'''
        function_reference = self.executor.instantiate_function_from_uuid(request.id,
                                           OpenOBDSession(request.session_info, grpc_factory=self.grpc_factory),
                                           OpenOBDFunctionBroker(self.arguments, grpc_factory=self.grpc_factory))

        self.executor.run_function(function_reference)

        return FunctionUpdate(
            type=FunctionUpdateType.FUNCTION_UPDATE_TYPE_RESPONSE,
            function_call=request,
            response=FunctionUpdateResponse.FUNCTION_UPDATE_SUCCESS,
            response_description="Run function locally"
        )

    def getFunctionBrokerToken(self, request, **kwargs):
        return FunctionBrokerToken(value="<none>")

    def generateFunctionSignature(self, request, **kwargs):
        return FunctionSignature(id=str(uuid.uuid4()),signature="<bogus>")

class NetworkGrpcFactoryBypassingFunctionBroker(NetworkGrpcFactory):

    def __init__(self, arguments: ConnectionArguments, executor: FunctionExecutor, grpc_host = None, grpc_port = 443):
        super().__init__(grpc_host=grpc_host, grpc_port=grpc_port)
        self.arguments = arguments
        self.executor = executor

    def get_function_broker(
            self) -> grpcFunctionBrokerServices.functionBrokerStub | grpcFunctionBrokerServices.functionBrokerServicer:
        return BypassFunctionBrokerServicer(self.arguments, self.executor, self)




import logging
import os
import typing

from openobd.core.openobd import OpenOBD
from openobd.core.function_broker import OpenOBDFunctionBroker
from openobd.core.grpc_factory import GrpcFactory, NetworkGrpcFactory
from openobd.core.session import OpenOBDSession
from openobd.core.arguments import ConnectionArguments, SessionArguments
from openobd_protocol.SessionController.Messages.SessionController_pb2 import SessionInfo
from openobd.functions.composition import OpenOBDComposition
from openobd.functions.function import OpenOBDFunction
from openobd.functions.bypass_function_broker import NetworkGrpcFactoryBypassingFunctionBroker
from openobd.functions.function_executor import FunctionExecutor


class SessionBuilder:
    session_arguments: SessionArguments
    connection_arguments: ConnectionArguments
    openobd_session = None
    function_broker = None

    def __init__(self, connection_arguments: ConnectionArguments, session_arguments: SessionArguments, function_executor: FunctionExecutor):
        self.connection_arguments = connection_arguments
        self.session_arguments = session_arguments

        self.function_executor = function_executor
        self.function_executor.load_modules(self._get_default_broker())

    def _create_grpc_factory(self, grpc_host, grpc_port=443) -> GrpcFactory:
        if self.session_arguments.bypass_function_broker:
            if self.function_executor is None:
                raise AssertionError("When bypassing function broker, you must specify a function executor to use")

            '''Return gRPC network factory that is bypassing the function broker calls'''
            return NetworkGrpcFactoryBypassingFunctionBroker(arguments=self.connection_arguments,
                                                             executor=self.function_executor,
                                                             grpc_host=grpc_host,
                                                             grpc_port=grpc_port)
        else:
            '''Return regular gRPC network factory'''
            return NetworkGrpcFactory(grpc_host=grpc_host, grpc_port=grpc_port)

    def _get_default_broker(self):
        return OpenOBDFunctionBroker(self.connection_arguments,
                              grpc_factory=self._create_grpc_factory(
                                  grpc_host=self.connection_arguments.grpc_host,
                                  grpc_port=self.connection_arguments.grpc_port)
                              )

    def _create_session(self) -> OpenOBDSession:
        if self.session_arguments.ticket_id is not None:
            return OpenOBD(self.connection_arguments).start_session_on_ticket(ticket_id=self.session_arguments.ticket_id)

        elif self.session_arguments.connector_id is not None:
            return OpenOBD(self.connection_arguments).start_session_on_connector(connector_id=self.session_arguments.connector_id)

        elif self.session_arguments.token is not None:
            session_info = SessionInfo("", "", "", self.connection_arguments.grpc_host, self.session_arguments.token)
            return OpenOBDSession(session_info)

        raise AssertionError("Please provide a ticket id (--ticket), connector id (--connector) or a token (--token) to initialize the session.")

    def _create(self):
        openobd_session = self._create_session()
        grpc_factory = self._create_grpc_factory(openobd_session.session_info.grpc_endpoint)
        self.openobd_session = OpenOBDSession(openobd_session.session_info, grpc_factory=grpc_factory)
        self.function_broker = OpenOBDFunctionBroker(self.connection_arguments, grpc_factory=grpc_factory)

    def function(self) -> OpenOBDFunction:
        self._create()
        return OpenOBDFunction(self.openobd_session, self.function_broker)

    def composition(self) -> OpenOBDComposition:
        self._create()
        return OpenOBDComposition(self.openobd_session, self.function_broker)

    def run(self, function: typing.Type[OpenOBDFunction] = None, file_name: typing.Optional[str] = None):
        '''Check if relative script path has been set, convert to module path'''
        if function is None and file_name is not None:
            file = self._script_file_to_module_path(file_name)
            if file is not None:
                '''
                Use the default broker instance using the provided arguments.
                This broker endpoint does not depend on the session info and therefore can be created
                before the openOBD session is created. Saving time in the case a function initialization fails.
                '''
                function = self.function_executor.load_function(file, self._get_default_broker())

        if function is None:
            logging.warning("Could not load function! No file reference given.")
            return

        '''First create the session and then run the function'''
        self._create()
        self.function_executor.run_function(function(self.openobd_session, self.function_broker))

    @staticmethod
    def _script_file_to_module_path(script):
        if script is None:
            return None
        return script.replace(os.sep, '.').replace('.py', '').lstrip(".")



from openobd.communication.response_exceptions import ResponseException
import logging

class Socket:

    def __init__(self, socket):
        self.socket_finished = False
        self.socket = socket
        self.__enter__()

    def __enter__(self):
        return self.socket

    def __del__(self):
        self.__exit__(None, None, None)

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            if isinstance(exc_type, ResponseException):
                logging.error(f'Request failed: {exc_value}')
            else:
                pass

        # Make sure we stop the socket once
        if self.socket_finished:
            return

        self.socket_finished = True

        # Finally stop the socket stream
        self.socket.stop_stream()




from openobd_protocol.Communication.Messages.Tp20_pb2 import Tp20Channel, Tp20Message

from openobd.communication.response_exceptions import *
from openobd.communication.uds_request_handler import UdsRequestHandler
from openobd.core.exceptions import OpenOBDStreamTimeoutException
from openobd.core.session import OpenOBDSession
from openobd.core.stream_handler import StreamHandler


class Tp20Socket:

    def __init__(self, openobd_session: OpenOBDSession, channel: Tp20Channel, timeout: float | None = 10):
        """
        Handles sending messages to the given channel and parsing incoming responses.

        :param openobd_session: an active OpenOBDSession with which to start the CAN stream.
        :param channel: the channel which to send messages to.
        :param timeout: time in seconds to wait for each response before raising a NoResponseException. None will wait forever.
        """
        self.stream_handler = StreamHandler(openobd_session.open_tp20_stream, outgoing_stream=True)
        self._channel = channel
        self.uds_request_handler = UdsRequestHandler(self.stream_handler, timeout)

        '''We send an empty message to the server to ensure that the channel is initialized, this also helps when we 
        want to just listen on a channel'''
        self.stream_handler.send(Tp20Message(channel=self._channel, payload=""))

    def request(self, payload: str, timeout: float | None = None, silent: bool = False, tries: int = 1) -> str | None:
        """
        Sends the given payload and waits for the response. Raises the appropriate ResponseException when receiving a
        negative response.

        :param payload: the bytes to send to the channel.
        :param timeout: time in seconds to wait for a response before raising a NoResponseException. Will use self.timeout if None.
        :param silent: if True, will not raise any ResponseExceptions. Instead, returns the payload of negative responses or None for NoResponseExceptions.
        :param tries: how many times to send the payload until a positive response is received.
        :return: the payload of the received response.
        """
        try:
            message = Tp20Message(channel=self._channel, payload=payload)
            return self.uds_request_handler.request(message, timeout, silent, tries)
        except ResponseException as e:
            # Add channel info to the exception before raising it
            e.request_id = self._channel.logical_address
            raise e

    def send(self, payload: str):
        message = Tp20Message(channel=self._channel, payload=payload)
        self.stream_handler.send(message, flush_incoming_messages=True)

    def receive(self, timeout: float):
        try:
            response = self.stream_handler.receive(timeout=timeout)     # type: Tp20Message
        except OpenOBDStreamTimeoutException:
            raise NoResponseException(request="??", response=None, request_id=self._channel.logical_address)

        return response.payload.upper()

    def stop_stream(self) -> None:
        """
        Closes the gRPC stream if it is not already closed. A new Socket object will have to be created to start another
        stream.
        """
        self.uds_request_handler.stop_stream()



import copy
import sys

from openobd.core.exceptions import OpenOBDException
from openobd.core.session import OpenOBDSession
from openobd.functions.function import OpenOBDFunction
from openobd.core.arguments import ConnectionArguments
from openobd_protocol.FunctionBroker.Messages.FunctionBroker_pb2 import FunctionUpdate, FunctionUpdateType, FunctionUpdateResponse, FunctionCall

from openobd.core.function_broker import OpenOBDFunctionBroker
from openobd.core.stream_handler import StreamHandler
from openobd.core.exceptions import OpenOBDStreamStoppedException

from openobd.functions.function_executor import FunctionExecutor

import threading
import logging

class FunctionLauncher:

    def __init__(self, connection_arguments: ConnectionArguments, function_executor: FunctionExecutor):
        """
        Handles communication with the function broker about functions that are available for execution.

        :param connection_arguments: Function launcher arguments
        :param function_executor: Functions that can be executed by the launcher
        """
        self.function_broker = OpenOBDFunctionBroker(connection_arguments)
        self.function_executor = function_executor
        self.function_executor.load_modules(self.function_broker)
        self.stream_handler = None
        self.process_function_updates_thread = None
        self.threads = []

    def serve(self):
        """
        Starts serving functions to the function broker
        """
        function_registrations = self.function_executor.get_function_registrations()
        if len(function_registrations) == 0:
            logging.critical("No functions to serve!")
            logging.critical("A function should be defined according to the following pattern:\n" + '''
class ECUReset(OpenOBDFunction):

    version = "v0.1"
    name = "ECU Reset"
    description = "Reset ECU"

    def run(self):
        logging.info("Reset ECU")                        

            ''')
            sys.exit(1)

        '''Establish function update stream'''
        self.stream_handler = StreamHandler(self.function_broker.open_function_stream, outgoing_stream=True)

        for function_registration in function_registrations:
            self.stream_handler.send(FunctionUpdate(type=FunctionUpdateType.FUNCTION_UPDATE_TYPE_REQUEST, function_registration=function_registration))

        self.process_function_updates_thread = threading.Thread(target=self._process_function_updates,
                                                                name="Process function updates",
                                                                daemon=True)
        self.process_function_updates_thread.start()

        try:
            self.process_function_updates_thread.join()
        except KeyboardInterrupt as e:
            self._interrupt_launcher()

    def _process_function_updates(self):
        try:
            while True:
                function_update = self.stream_handler.receive()
                self._function_update(function_update)
        except OpenOBDStreamStoppedException:
            pass
        except OpenOBDException as e:
            logging.error(f"Stopped processing function updates due to an exception.")
            logging.error(e)
        finally:
            self.stream_handler.stop_stream()

    def _interrupt_launcher(self):
        for (function_reference, function_thread) in self.threads:
            if function_reference.__is_active__():
                logging.info(f"Stopping {function_reference}...")
                function_reference.interrupt()

        self.stream_handler.stop_stream()

    def _function_update(self, function_update: FunctionUpdate):
        if function_update.type == FunctionUpdateType.FUNCTION_UPDATE_TYPE_REQUEST:
            if function_update.HasField('function_call'):
                call = function_update.function_call
                logging.info(f"Received a function call for function id [{function_update.function_call.id}]")

                try:
                    # Test if function is available
                    self._available(call)

                    function_reference = self._get_function_instance(call)

                    function_thread = threading.Thread(target=self._call,
                                                     name=f"Function [{function_update.function_call.id}]",
                                                     args=[function_reference],
                                                     daemon=True)

                    self.threads.append((function_reference, function_thread))
                    function_thread.start()

                    # Immediately report back that we successfully received the function call and started the function
                    self.stream_handler.send(
                        FunctionUpdate(
                            type=FunctionUpdateType.FUNCTION_UPDATE_TYPE_RESPONSE,
                            function_call=function_update.function_call,
                            response=FunctionUpdateResponse.FUNCTION_UPDATE_SUCCESS
                        ))

                except Exception as e:
                    self.stream_handler.send(
                        FunctionUpdate(
                            type=FunctionUpdateType.FUNCTION_UPDATE_TYPE_RESPONSE,
                            function_call=function_update.function_call,
                            response=FunctionUpdateResponse.FUNCTION_UPDATE_FAILED,
                            response_description=str(e)
                        ))

            elif function_update.HasField('function_registration'):
                raise OpenOBDException("We do not expect a 'function_registration' as request from the function broker.")

            elif function_update.HasField('function_broker_token'):
                logging.info(f"Function update stream is still open")
                logging.debug(f"Received a new function broker token")

                # Acknowledge the ping/token update
                # We need to communicate to the broker every so often (before the Connection idle timeout of the Load Balancer)
                # or the stream (and all functions we host) gets closed
                response = copy.copy(function_update)
                response.type=FunctionUpdateType.FUNCTION_UPDATE_TYPE_RESPONSE
                self.stream_handler.send(response)

                logging.debug(f"Send new function broker token response")

            elif function_update.HasField('function_broker_reconnect'):
                raise OpenOBDException("Received a reconnect request from the function broker. We do not support it yet.")

            else:
                raise OpenOBDException("Unsupported request from the function broker")

        elif function_update.type == FunctionUpdateType.FUNCTION_UPDATE_TYPE_RESPONSE:
            logging.debug("Received function update response...")

            if function_update.HasField('function_registration') and function_update.HasField('response'):
                logging.debug(f"Received a function update for function id [{function_update.function_registration.details.id}]")

                if function_update.response == FunctionUpdateResponse.FUNCTION_UPDATE_SUCCESS:
                    logging.info(f"Function [{function_update.function_registration.details.id}] registration was successful")
                elif function_update.response == FunctionUpdateResponse.FUNCTION_UPDATE_UNAUTHORIZED:
                    logging.warning(f"Function [{function_update.function_registration.details.id}] signature check failed!")
                else:
                    logging.warning(f"Function [{function_update.function_registration.details.id}] registration failed!")


    def _get_function_instance(self, call: FunctionCall):
        return self.function_executor.instantiate_function_from_uuid(id=call.id,
                                                                     openobd_session=OpenOBDSession(call.session_info),
                                                                     function_broker=self.function_broker)

    def _call(self, function: OpenOBDFunction):
        self.function_executor.run_function(function)

    def _available(self, call: FunctionCall):
        self.function_executor.instantiate_function_from_uuid(id=call.id,
                                                              openobd_session=OpenOBDSession(call.session_info),
                                                              function_broker=self.function_broker,
                                                              dry_run=True)







import logging

from openobd_protocol.FunctionBroker.Messages.FunctionBroker_pb2 import FunctionRegistration, FunctionRegistrationState
from openobd_protocol.Function.Messages.Function_pb2 import FunctionDetails
from openobd_protocol.Messages.Empty_pb2 import EmptyMessage
from openobd_protocol.Session.Messages.ServiceResult_pb2 import ServiceResult, Result
from openobd.core.exceptions import OpenOBDException
from openobd.core.session_token_handler import SessionTokenHandler
from openobd.core.stream_handler import StreamHandler
from openobd.core.session import OpenOBDSession
from openobd.core.function_broker import OpenOBDFunctionBroker
from openobd.functions.function_context_handler import FunctionContextHandler

class OpenOBDFunction:
    id = ""
    version = "<undef>"
    name = ""
    description = ""
    signature = ""
    openobd_session = None

    '''
    Optionally you can pass a OpenOBDFunctionBroker (when non-standard configuration is needed)
    '''
    def __init__(self, openobd_session: OpenOBDSession, function_broker: OpenOBDFunctionBroker):
        try:
            self.openobd_session = openobd_session
            self.function_broker = function_broker
            self.result = ServiceResult(result=[Result.RESULT_SUCCESS])
            self._init_function_details()
            self._function_registration = FunctionRegistration(
                details=self._function_details,
                state=FunctionRegistrationState.FUNCTION_REGISTRATION_STATE_ONLINE,
                signature=self.signature)
            self._context_finished_ = False
        except OpenOBDException as e:
            logging.error(f"Failed to initialize openOBD function: {e}")
            self._function_registration = None
            self._context_finished_ = True

    def run(self):
        pass

    def get_function_registration(self):
        return self._function_registration

    def get_function_id(self):
        return self.id

    def _init_function_details(self, id=None, version=None, name=None, description=None, signature=None):
        self._function_details = None

        '''Check id presence, if not present we generate a fresh one'''
        if not id is None:
            self.id = id
        elif len(self.id) == 0:
            function = self.function_broker.generate_function_signature()
            self.id = function.id
            self.signature = function.signature

        '''Check signature presence'''
        if not signature is None:
            self.signature = signature
        elif len(self.signature) == 0:
            raise OpenOBDException("A function signature is required in order to serve a function on the network!")

        '''Initialize (version, name, description) correctly'''
        if not version is None:
            self.version = version

        '''Only overwrite name with class name when no custom name is provided'''
        if not name is None:
            self.name = name
        elif len(self.name) == 0:
            self.name = self.__class__.__name__

        if not description is None:
            self.description = description

        self._function_details = FunctionDetails(id=self.id,
                                                 version=self.version, name=self.name, description=self.description)

    def __enter__(self):
        if self._context_finished_:
            ''' We cannot continue when the constructor failed '''
            raise OpenOBDException("Failed to construct openOBD session")

        self._authenticate()
        return self

    # Authenticate against the openOBD session for this function
    def _authenticate(self):
        try:
            # Start the SessionTokenHandler to ensure the openOBD session remains authenticated
            SessionTokenHandler(self.openobd_session)
        except OpenOBDException as e:
            logging.error(f'Activating openOBD session failed: {e}')
            self._context_finished_ = True
            raise

    def __is_active__(self):
        return not self._context_finished_

    def __del__(self):
        self.__exit__(None, None, None)

    def interrupt(self):
        self.__exit__(OpenOBDException(f"Interrupt"), f"Function [{self.id}] has been interrupted!", None)

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            self.result = ServiceResult(result=[Result.RESULT_FAILURE])
            if isinstance(exc_type, OpenOBDException):
                logging.error(f'Request failed: {exc_value}')
            else:
                logging.error(f"Something unexpected occurred [{exc_value}]")
                # if traceback is not None:
                #     logging.warning(traceback.format_exc())

        # Make sure we finish the context once
        if not hasattr(self, '_context_finished_'):
            # Something wrong with initialization, also no cleanup possible
            return

        if self._context_finished_:
            return

        self._context_finished_ = True
        if self.openobd_session is not None:
            # Check if result is set
            if not hasattr(self, "result"):
                self.result = ServiceResult(result=[Result.RESULT_FAILURE])

            response = self.openobd_session.finish(self.result)
            self.openobd_session = None
            logging.info(response)

    def set_bus_configuration(self, bus_configs):
        # Open a configureBus stream, send the bus configurations, and close the stream
        bus_config_stream = StreamHandler(self.openobd_session.configure_bus)
        bus_config_stream.send_and_close(bus_configs)
        logging.info("Buses have been configured.")

    def run_function(self, function_class):
        '''Function id detection'''
        if issubclass(function_class, OpenOBDFunction):
            function_id = function_class.id
        else:
            raise OpenOBDException(f"Function not found: {function_class.name}")

        function_context = FunctionContextHandler(self.openobd_session)
        function_context.run_function(function_id, self.function_broker)
        return self.get_results()

    def set_result(self, key, value):
        self.openobd_session.set_function_result(key=key, value=value)

        # This is the wrong way?
        # self.openobd_session.function.setFunctionResult(Variable(type=ContextType.FUNCTION_CONTEXT, key=key, value=value))

    def set_variable(self, key, value):
        self.openobd_session.set_context_variable(key=key, value=value)

    '''Return the results as a dictionary'''
    def get_results(self) -> dict:
        function = self.openobd_session.function
        result_list = function.getFunctionResultList(request=EmptyMessage(),
                                                     metadata=self.openobd_session._metadata())  # type: VariableList
        variables = {}
        for var in result_list.variables:
            variables[var.key] = var.value
        return variables
